/* *** VM1Plc.ino - Basic support software for VM1AT_R1 board ***

 This software contains a simple example of a soft PLC with a MODBUS
 RTU server via serial link.
 
 The functions included are:
	- Two task dispatchers with 10ms and 100ms periods
	- Basic PLC like operation with 10 ms scan cycle.
	- Modbus Slave RTU using software serial link	
	- Use examples of VM1ATR1 support library
 
 The library ModbusServer is based in parts of the Modbus library developed by
 Jason Vreeland [CodeRage] on 3/7/2010. Thank you Jason for your insights.
 This implementation of the Modbus Server uses a custom defined Data Table with
 both Boolean (Digital Inputs, Coils) and Analog(Input and Holding Registers) arrays.
 The modbus function codes implemented are:

*********************************************************************
 LICENSE

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********************************************************************
*/
#include <SoftwareSerial.h>		//Used by Modbus Server library
#include "ModbusServer.h"		//Modbus server library
#include "MemoryFree.h"			//Used to check how much RAM is left
#include "vm1atr1.h"			//Support library
#include "ModbusTags.h"			//Modbus table enumerations

// *** Some propaganda ***
#define PLC_VERSION 10100	//Version 1.0.0 - Change before compilation
#define ver1 "1"
#define ver2 "1"
#define ver3 "0"
#define ver123 ver1"."ver2"."ver3
#pragma GCC diagnostic ignored "-Wwrite-strings"
const char* sversion = ver123"-"__DATE__"-"__TIME__;
const char* company = "TUTAN_AUTOMATION_CANADA";
const char* mcu_type = "ATMEGA328P";

//*** DEBUG PRINT PREFERENCES ***
#define DEBUG 		//Allow debugging via serial monitor. Comment out to disable
//Next line are used to use printf() instead of Serial.print() for the serial monitor
#define USE_PRINTF	//Comment out to disable

// *** HARDWARE CONSTANTS - Cannot be modified ***
const uint8_t blinkerPin = 9;	//STATUS LED
const uint8_t tx1Pin = 7;		//Serial transmit pin (TB3-2) TX LED
const uint8_t rx1Pin = 8;		//Serial receive pin (TB3-3) RX LED
const int VM1ATR1 = 1;

// *** Aplication constants - Can be modified to suit your requirements

// --- MODBUS LINK PARAMETERS ---
const uint8_t deviceID = 1;			//This is the ID for this Modbus Server, can be set from 1 to 255
const uint32_t baudRate = 38400;	//Serial link baud rate can be set from 300 to 115200 baud

// Modbus data table size - Be careful with this numbers to avoid running our of memory!
const uint16_t anaTblLen = 128;		//Number of analog registers in modbus data table
const uint16_t booTblLen = 128;		//Number of boolean values in modbus data table

//Create the MODBUS data table
static byte coils[booTblLen / 8 + 1];   //Modbus coils (each byte holds 8 coils
static int hregs[anaTblLen];      		//Modbus hregs (each int holds 1 register

//Create instance of MODBUS server - Uses SoftwareSerial library
ModbusServer mbs(deviceID, rx1Pin, tx1Pin, baudRate, anaTblLen, booTblLen, hregs, coils);

//Event counters
uint32_t eventCounter10ms = 0;		//Counts the number of executions of task10ms function
uint32_t eventCounter100ms = 0;		//Counts the number of executions of task100ms function

//Set up stdout to enable printf function via serial link
#ifdef USE_PRINTF
FILE serial_stdout;
 int serial_putchar(char c, FILE* f) {
    if (c == '\n') serial_putchar('\r', f);
    return Serial.write(c) == 1? 0 : 1;
}
#endif
/* ************************************************************
	*** MODUS DATA TABLE ACCESS HELPER FUNCTIONS ***
	
	This functions perform the basic read/write operations to the
	MODBUS data table.
	Functions:
	
************************************************************** */
//Get one boolean value from modbus coils
bool getBoo(uint16_t index)
{
	return (bitRead(coils[index / 8], index % 8));
}
 //Set one boolean value into ModBus coils
void setBoo (uint16_t index, bool value)
{
    bitWrite(coils[index/8], index % 8, value);
}
//Get one integer from modbus register
int getAna(uint16_t index)
{
	return(hregs[index]);
}
//Set one integer into modbus register
void setAna(uint16_t index, int32_t value)
{
	hregs[index] = value;
}
//Get one long unsigned integer from two modbus registers
uint32_t getDword(uint16_t index)
{
        return( (uint32_t)((uint16_t)hregs[index] | ((uint32_t)hregs[index+1] << 16) & 0xFFFF0000));
}
//Set one long unsigned integer into two modbus registers
void setDword(uint16_t index, uint32_t value)
{
        hregs[index] = (uint16_t)(value & 0xFFFF);
        hregs[index + 1] = (uint16_t)((value >> 16) & 0xFFFF);
}
/* ************************************************************
	*** PLC I/O functions - EXAMPLE ***
	
	This functions perform the high level functions
	to update inputs and outputs to the PLC logic
	In this example all values are stored/retrieved from the 
	Modbus datatable, allowing remote access to them.
	
	INPUTS: Can be read from Modbus Client or local functio0ns.
	Writes will be overwritten by PLC read functions.
	
	OUTPUTS: Can be written from Modbus Client or local functions. 
	All remote reads will return the last value written either by the
	Modbus client or the local apllication.
	
************************************************************** */
// Read digital inputs into modbus Data Table
void updateDigitalInputs(void)
{
    uint16_t dinput;
    dinput = vm.readDigitalInputs();  			//Read 16 digital inputs
	//Transfer values to the modbus data table 8 bits a a time
    coils[X_DINP00/8] = dinput & 0xff;  		//Set LSB in DT beginning at DINP00 (DINP00 to DINP07)
    coils[X_DINP08/8] = (dinput >> 8) & 0xff;	//Set MSB in DT beginning at DINP08 (DINP08 to DINP15)
}
//Write digital outputs from modbus Data table
void updateDigitalOutputs(void)        //Write digital outputs from Data table
{
    uint8_t doutput;
    doutput = coils[Q_COIL00/8];       	//Get value 8 bits together
    vm.writeDigitalOutputs(doutput);   	//Update outputs
 }
uint8_t readBackDigitalOutputs(void)
{
	return coils[Q_COIL00/8];
}	
	//Read analog inputs and update Data Table
void updateAnalogInputs()
{
	for (int i=0; i<=5; i++)
	{
		hregs[I_AINP00 + i] = vm.readAnalogInput(i);	//Set in modbus data table
	}
}
//write analog outputs from Data Table
void updateAnalogOutputs(void)
{
	for (int i=0; i<=1; i++)
	{
		vm.writeAnalogOutput(O_AOUT00+i, getAna(O_AOUT00+i));
	}
}
/* ************************************************************
	*** PLC operational functions ***
	Called in sequence from plcOperationProcess()
	Functions:
	plcOP_01();	//Do something
	plcOP_02();	//etc..
************************************************************** */
void plcOP_01(void)
{
	//Example: Turn ON digital output 0 when both digital inputs 0 AND 1 are ON
	setBoo(Q_COIL00, getBoo(X_DINP00) && getBoo(X_DINP01));
	
	//Example: Set three outputs as LOW , OK and HIGH limits of analog input 4
	const int vMin = 100;	//Min. value
	const int vMax = 900; 	//Max. value
	int value = getAna(I_AINP04); //Read analog input 4 - TB4(3)
	setBoo(Q_AINPLO, value < vMin); 					//LOW
	setBoo(Q_AINPOK, value >= vMin && value <= vMax); 	//OK
	setBoo(Q_AINPHI, value > vMax); 					//HIGH
}
void plcOP_02(void)
{
	//Another sub process here...
}	
/* ************************************************************
	*** PLC SCAN PROCESSES ***
		
	The three functions below perform the basic PLC operation
	in order:
		1 - Read all inputs
		2 - Process logic as required
		3 - Write all outputs
	Place your calls to specific functions inside the plcOperationProcess() function

	This is only one example of how to create a PLC-like behaviour
	with code and does not implies that is the best procedure for
	your specifica application.
	
************************************************************** */

void plcInputProcess(){
	updateDigitalInputs();
	updateAnalogInputs();
}
void plcOperationProcess(){
	//Here comes your PLC logic.
	//Usually only calls to simple processes are placed here
	//to keep code clean and easy to maintain.
	
	plcOP_01();	//Do something
	plcOP_02();	//etc..
	//etc..
	
	
}
void plcOutputProcess(){
	updateDigitalOutputs();	
	updateAnalogOutputs();
}
/* ************************************************************
	*** TASK DISPATCHERS ***
	
	In this example there are two task dispatchers running every
	10 and 100 ms respectively.
	Inside each dispatcher there is a switch case statement with 10
	cases. Functions called from each case are executed only once every
	10 calls to the dispatcher (i.e. 100ms and 1000ms). Functions called
	outside the case statement are executed at every scan.
	This approach allows to distribute the calls in an orderly manner to
	give each function enough time to complete before calling the next one.

	Use the 10ms task to call high speed functions that require
	fast monitoring.

************************************************************** */
void task10ms(uint32_t time)
{
	static byte index = 0;
    eventCounter10ms++;
    mbs.run();			//Run the Modbus Server
	switch(index++)
	{
		case (0):	//Executed at 0, 100, 200, 300 ... 1100, 1200, 1300 ms etc..
		{
			plcInputProcess();
 			break;	}
		case (1):   //Executed at 10, 110, 210, 310  ... 1110, 1210, 1310 ms etc..
		{	plcOperationProcess();
			break;	}
		case (2):   //Executed at 20, 120, 220, 320  ... 1120, 1220, 1320 ms etc..
		{	plcOutputProcess();
			break;	}
		case (3):   //Executed at 30, 130, 230, 330  ... 1130, 1230, 1330 ms etc..
		{
			break;	}
		case (4):   //Executed at 40, 140, 240, 340  ... 1140, 1240, 1340 ms etc..   
		{
			break;	}
		case (5):   //Executed at 50, 150, 250, 350  ... 1150, 1250, 1350 ms etc..   
		{
			break;	}
		case (6):   //Executed at 60, 160, 260, 360  ... 1160, 1260, 1360 ms etc..   
		{
			break;  }
		case (7):   //Executed at 70, 170, 270, 370  ... 1170, 1270, 1370 ms etc..   
		{
		    break;	}
		case (8):   //Executed at 80, 180, 280, 380  ... 1180, 1280, 1380 ms etc..   
		{
			break;	}
		case (9):   //Executed at 90, 190, 290, 390  ... 1190, 1290, 1390 ms etc..   
		{
			break;	}
	}
	if (index > 9) index = 0;
}

void task100ms(uint32_t time)
{
    static byte index = 0;
    eventCounter100ms++;

	switch(index++)
	{
		case (0):	//Executed at 0, 1000, 2000, 3000 ... 11000, 12000, 13000 ms etc..   
		{
			break;	}
		case (1):   //Executed at 100, 1100, 2100, 3100 ... 11100, 12100, 13100 ms etc..   
		{
			break;	}
		case (2):   //Executed at 200, 1200, 2200, 3200 ... 11200, 12200, 13200 ms etc..   
		{
			break;	}
		case (3):   //Executed at 300, 1300, 2300, 3300 ... 11300, 12300, 13300 ms etc..   
		{
            break;  }
        case (4):   //Executed at 400, 1400, 2400, 3400 ... 11400, 12400, 13400 ms etc..   
		{
            break;	}
		case (5):   //Executed at 500, 1500, 2500, 3500 ... 11500, 12500, 13500 ms etc..   
		{
			break;	}
		case (6):   //Executed at 600, 1600, 2600, 3600 ... 11600, 12600, 13600 ms etc..   
		{
			break;	}
		case (7):   //Executed at 700, 1700, 2700, 3700 ... 11700, 12700, 13700 ms etc..   
		{
			//Update event counters in modbus data table
			setDword(D_T010LO, eventCounter10ms);
			setDword(D_T100LO, eventCounter100ms);
		    break;	}
		case (8):   //Executed at 800, 1800, 2800, 3800 ... 11800, 12800, 13800 ms etc..   
		{
			int i;
			//Example: Send some activity to serial monitor once per second
			#ifdef DEBUG
			#ifdef USE_PRINTF
			printf("Ainp=");
			for (i=0; i<=5; i++)
			{
				printf("%04d,",getAna(I_AINP00+i));
			}
			printf(" Dinp=");
			for (i=0; i<=16; i++)
			{ 
				printf("%d,",getBoo(i));	//Print all digital inputs
			}
			printf(" Dout= ");
			for (i = 0; i<=7; i++)
			{ 
				printf("%d,",getBoo(i+Q_COIL00));	//Print all digital outputs
			}
			printf("\n");
			#endif
			#endif
			break;	}
		case (9):   //Executed at 900, 1900, 2900, 3900 ... 11900, 12900, 13900 ms etc..   
		{
			digitalWrite(blinkerPin, ~digitalRead(blinkerPin) & 1); //Toggle the blinker every second
			break;	}
	}
	if (index > 9) index = 0;	
}

//The setup function is called once at startup of the sketch
void setup()
{
	//Serial link used to monitor operations using the FTDI connection
	Serial.begin(9600);	//You may select a different baud rate up to 115200
	//Set up stdout to enable printf function via serial link
	//The next two lines are needed only if you prefer to use printf() instead of Serial.print()
	#ifdef USE_PRINTF
	fdev_setup_stream(&serial_stdout, serial_putchar, NULL, _FDEV_SETUP_WRITE);
    stdout = &serial_stdout;
	#endif
	
	#ifdef USE_PRINTF
	//Splash messages - Modify and use them if you like the idea
	printf("VM1AT_R1 - Industrial Micro-controller - MCU : %s\n", mcu_type);
	printf("Version: %s\n",sversion);
	printf("by: %s\n",company);
	printf("FreeMemory = %d bytes\n", freeMemory());
	printf("Boolean Table: %d coils\n", booTblLen);
	printf("Analog  Table: %d hregs\n", anaTblLen);
	#endif

	//Set I/O pins modes
	pinMode(blinkerPin, OUTPUT);	//Set blinker pin to output mode
	//analogReference(INTERNAL);		//Set analog reference to 1.1 Volt (for use with LM35 sensor)
}

// The loop function is used to call the 10ms and 100ms task
// that dispatch all PLC functions.
void loop()
{
    static uint32_t t10Old, t100Old;
    uint32_t t10New, t100New;
    t10New = millis();
    if ((long)(t10New - t10Old) >= 10) {
        t10Old = t10New;
        task10ms(t10New);   //Execute 10ms task
    }
    t100New = millis();
    if ((long)(t100New - t100Old) >= 100) {
        t100Old = t100New;
        task100ms(t100New); //Execute 100ms task
    }
}

