/* *** VMbasic.ino - Simple example for VM1AT_R1 board ***

 This software contains simple examples to access the basic functionality
 of the board with explanatory comments that may help to develop other tasks.
 The low level access to all I/O is performed by the library VM1ATR1, which
 is supplied with the board.
 All software related to this board can be downloaded at:
 www.tutanautomation.com/downloads
 
 The functions include:
        - Use the board's serial link
	- Read 16 Digital inputs
        - Read 6 Analog inputs
        - Write 7 Digital Outputs as requested by serial commands

*********************************************************************
 LICENSE

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ********************************************************************
*/
#include <SoftwareSerial.h> //Serial link library
#include "VM1ATR1.h"  	    //Board support library

const uint8_t blinkerPin = 9;	//STATUS LED
const uint8_t tx1Pin = 7;	//Serial transmit pin (TB3-2) TX LED
const uint8_t rx1Pin = 8;	//Serial receive pin (TB3-3) RX LED

uint32_t lastTime = 0;         //Variable to store the time

SoftwareSerial ss(rx1Pin, tx1Pin);  //Create instance of the serial library

void setup()
{
	Serial.begin(9600);    //Debug link - You may use a different speed
        ss.begin(9600);        //Board link - You may use a different speed
	pinMode(blinkerPin, OUTPUT);  //Set the blinker pin as an output
	lastTime = millis();  //Init the last time variable
}
//
//  This sketch checks continuously for charactters recived through the serial links and
//  toggles one digital output according to the character received. 
//  Every second, toggles the blinker LED and sends all I/O data through the two serial channels.
//
void loop()
{
    uint32_t timeNow = millis();
    
    // Check for characters received by any serial link
    checkSerial_0();   //Process characters received via debug link (RX0/TX0)
    checkSerial_1();   //Process characters received via VM1ATR1 serial link (RX1/TX1)
    
    //Using the millis() function instead of delay() we may keep checking the serial
    //link every loop while doing other things only once a second.
    if ((long)(timeNow - lastTime) >= 1000) 
	{
		lastTime = timeNow;
		digitalWrite(blinkerPin, ~digitalRead(blinkerPin) & 1); //Toggle the blinker 
		monitor_0();	//Send all I/O data via debug link (RX0/TX0)
                monitor_1();     //Send all I/O data via VM1ATR1 serial link (RX1/TX1)
	}

}
//
//  This function checks for characters received through debug link (RX0/TX0)
//  and toggles a Digital output correcponding the to characeter recived.
//  Character "0" to "6" toggle Digital Outputs 0 to 6
//
void checkSerial_0()
{
	char c;
	uint8_t index;
    if (Serial.available()) 
	{
		c = Serial.read();
		index = c - '0';
		if (index < 0) index = 0;
                if (index > 6) index - 6;
		vm.writeDigitalOutput(index, ~vm.readDigitalOutput(index) & 1); //Toggle output
	}
}
//
//  This function checks for characters received through VM1ATR1 seial link (RX1/TX1)
//  and toggles a Digital output correcponding the to characeter recived.
//  Character "0" to "6" toggle Digital Outputs 0 to 6
//
void checkSerial_1()
{
	char c;
	uint8_t index;
	if(ss.available())
		{
		c = ss.read();
		index = c - '0';
		if (index < 0) index = 0;
                if (index > 6) index - 6;
		vm.writeDigitalOutput(index, ~vm.readDigitalOutput(index) & 1); //Toggle output
	}	
}	
//
//  This functions send all I/O values to serial link RX0/TX0 (donwload/debug link)
//  available at connector P1
//
void monitor_0()
{
	int i;

        //Hardware serial - Uses download/debug link - Connector P4  (Rx0 -> P1(4), TX0 -> P1(5))
        // First send all Analog Inputs
	Serial.print("Ainp=");
	for (i=0; i<=5; i++)
	{
		Serial.print(vm.readAnalogInput(i),DEC);
		Serial.print(",");
	}
        // Now send all Digital Inputs
	Serial.print(" Dinp=");
	for (i=0; i<=16; i++)
	{ 
		Serial.print(vm.readDigitalInput(i));	//Print all digital inputs
		Serial.print(",");
	}
        // Finally all Digital Outputs
	Serial.print(" Dout= ");
	for (i = 0; i<=6; i++)
	{ 
		Serial.print(vm.readDigitalOutput(i));
		Serial.print(",");
	}
	
	Serial.println();
}
//
//  This functions send all I/O values to serial link RX1/TX1
//  available at terminal block TB3
//
void monitor_1()
{
	int i;

 	// Software serial - Uses VM1ATR1 link: Connector TB3 (RX1 -> TB3(3), TX1 -> TB3(2))
	ss.print("Ainp=");
	for (i=0; i<=5; i++)
	{
		ss.print(vm.readAnalogInput(i));
		ss.print(",");
	}
	ss.println();	
	ss.print("Dinp=");
	for (i=0; i<=16; i++)
	{ 
		ss.print(vm.readDigitalInput(i));	//Print all digital inputs
		ss.print(",");
	}
	ss.println();	
	ss.print("Dout= ");
	for (i = 0; i<=6; i++)
	{ 
		ss.print(vm.readDigitalOutput(i));
		ss.print(",");
	}
	ss.println();	
}
